# Arch Perl library, Copyright (C) 2004 Mikhael Goikhman
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use 5.005;
use strict;

package Arch::Session;

use base 'Arch::Storage';

use Arch::Util qw(run_tla standardize_date parse_creator_email);
use Arch::TempFiles qw(temp_dir_name);
use Arch::Changeset;
use Arch::Log;

sub new ($%) {
	my $class = shift;
	my %init = @_;
	return $class->SUPER::new(%init);
}

sub working_names ($;$$$$$) {
	my $self = shift;
	if (@_) {
		my @names = $self->SUPER::working_names;
		$self->{categories} = undef
			unless defined $names[0] && defined $_[0] && $names[0] eq $_[0];
		$self->{branches}   = undef
			unless defined $names[1] && defined $_[1] && $names[1] eq $_[1];
		$self->{versions}   = undef
			unless defined $names[2] && defined $_[2] && $names[2] eq $_[2];
		$self->{revisions}  = $self->{revision_details} = undef
			unless defined $names[3] && defined $_[3] && $names[3] eq $_[3];
	}
	return $self->SUPER::working_names(@_);
}

sub archives ($) { 
	my $self = shift;
	$self->{archives} ||= [ run_tla("archives -n") ];
	return $self->{archives};
}
 
sub is_archive_registered ($;$) {
	my $self = shift;
	my $archive = @_? shift: $self->{archive};
	die "No working archive" unless defined $archive;

	unless ($self->{archives_presence}) {
		my $archives_hash = {};
		$archives_hash->{$_} = 1 foreach @{$self->archives};
		$self->{archives_presence} = $archives_hash;
	}
	return $self->{archives_presence}->{$archive};
}

sub categories ($;$) {
	my $self = shift;
	my $archive  = @_ > 0? shift: $self->{archive};
	die "No working archive" unless defined $archive;

	$self->{categories} ||= [
		map { s/^  //; $_ } grep { /^  / }
			run_tla("abrowse --categories $archive")
	];
	return $self->{categories};
}

sub branches ($;$$) {
	my $self = shift;
	my $archive  = @_ > 1? shift: $self->{archive};
	my $category = @_ > 0? shift: $self->{category};
	die "No working archive" unless defined $archive;
	die "No working category" unless defined $category;

	$self->{branches} ||= [
		map { s/^    //; $_ = "$category--" if $_ eq $category; $_ } grep { /^    / }
			run_tla("abrowse --branches", $self->working_name)
	];
	return $self->{branches};
}

sub versions ($;$$$) {
	my $self = shift;
	my $archive  = @_ > 2? shift: $self->{archive};
	my $category = @_ > 1? shift: $self->{category};
	my $branch   = @_ > 0? shift: $self->{branch};
	$branch = "" unless defined $branch;  # support branchless revisions
	die "No working archive"  unless defined $archive;
	die "No working category" unless defined $category;
	die "No working branch"   unless defined $branch;

	$self->{versions} ||= [
		map { s/^      //; $_ = "" if $_ eq $branch; $_ } grep { /^      / }
			run_tla("abrowse --versions", $self->working_name)
	];
	$self->{versions} = [ map { s/--/----/; $_ } grep !/--.*--/, @{$self->{versions}} ]
		if $branch eq '';
	return $self->{versions};
}

sub revisions ($;$$$$) {
	my $self = shift;
	my $archive  = @_ > 3? shift: $self->{archive};
	my $category = @_ > 2? shift: $self->{category};
	my $branch   = @_ > 1? shift: $self->{branch};
	my $version  = @_ > 0? shift: $self->{version};
	die "No working archive"  unless defined $archive;
	die "No working category" unless defined $category;
	die "No working branch"   unless defined $branch;
	die "No working version"  unless defined $version;

	unless ($self->{revisions}) {
		$self->{revisions} = [ run_tla("revisions", $self->working_name) ];
	}
	return $self->{revisions};
}

sub revision_details ($;$$$$) {
	my $self = shift;
	my $archive  = @_ > 3? shift: $self->{archive};
	my $category = @_ > 2? shift: $self->{category};
	my $branch   = @_ > 1? shift: $self->{branch};
	my $version  = @_ > 0? shift: $self->{version};
	die "No working archive"  unless defined $archive;
	die "No working category" unless defined $category;
	die "No working branch"   unless defined $branch;
	die "No working version"  unless defined $version;

	unless ($self->{revision_details}) {
		my $full_version = $self->working_name;
		my $version = $full_version;
		$version =~ s|^.*/||;

		# $ok is used to work around the tla bug with branchless version
		# $prev_line is used to track revisions with no (empty) summary
		my $ok = 0;
		my $prev_line = "";

		my @revision_lines = map { s/^        //? $_: undef }
			grep {
				$ok = /^      \Q$version\E$/ if /^      [^ ]/;
				my $end = ($prev_line =~ /^        /) && ($_ eq "");
				$prev_line = $_;
				($end || /^        /) && $ok
			}
			run_tla("abrowse --desc", $full_version);

		my @revision_details = ();
		while (@revision_lines) {
			my ($line1, $line2) = splice @revision_lines, 0, 2;
			my @summary_lines = ();
			push @summary_lines, shift @revision_lines while defined $revision_lines[0];
			shift @revision_lines;  # throw away undef delimiter
			my $summary = join("\n", @summary_lines);
			$line2 =~ s/^  //; $summary =~ s/^  //mg;

			my ($name, $kind) = $line1 =~ /^(\S+)\s+\((.*?)\)/
				or die "Unexpected output of tla abrowse --desc, subline 1:\n\t$line1\n";
			$kind = $kind =~ /tag/? "tag": $kind =~ /import/? "import": "cset";
			my ($date, $creator) = $line2 =~ /^(.+?)\s{6}(.*)/
				or die "Unexpected output of tla abrowse --desc, subline 2:\n\t$line2\n";
			$date = standardize_date($date);

			my ($creator_name, $creator_email) = parse_creator_email($creator);
			push @revision_details, {
				name    => $name,
				summary => $summary,
				creator => $creator_name,
				email   => $creator_email,
				date    => $date,
				kind    => $kind,
			};
		}
		$self->{revision_details} = \@revision_details;
		$self->{revisions} = [ map { $_->{name} } @revision_details ];
	}
	return $self->{revision_details};
}

# [
#   [ category1, [
#     [ branch1, [
#       [ version1, start_revision1, end_revision1 ],
#       [ version2, start_revision2, end_revision2 ],
#     ] ],
#     [ branch2, [
#       [ version3, start_revision3, end_revision3 ],
#       [ version4, start_revision4, end_revision4 ],
#     ] ],
#     ...,
#   ] ],
# ]

sub expanded_archive_info ($;$$) {
	my $self = shift;
	die "expanded_archive_info: no working archive\n" unless defined $self->{archive};
	my $archive_name = $self->working_name || shift;
	my $full_listing = shift || 0;  # currently ignored

	my $infos = [];
	my @category_infos = split(/^\b/m, join('',
		map { s/^  //; "$_\n" } grep { /^  / }
			run_tla("abrowse $archive_name")
	));

	my $error = 0;
	CATEGORY_ITEM:
	foreach (@category_infos) {
		my ($category, $branch_infos) = /^([^\s]+)\n(  .*)$/s;
		push @$infos, [ $category, [] ];
		unless (defined $category) {
			$error = 1; next CATEGORY_ITEM;
		}

		my @branch_infos = split(/^\b/m, join('',
			map { s/^  // or $error = 1; "$_\n" }
				split("\n", $branch_infos)
		));
		$error = 1 unless @branch_infos;
		foreach (@branch_infos) {
			my ($branch, $version_infos) = /^\Q$category\E(?:--([^\s]+))?\n(  .*)$/s;
			$branch = "" if defined $version_infos && !defined $branch;
			unless (defined $branch) {
				$error = 1; next CATEGORY_ITEM;
			}
			push @{$infos->[-1]->[1]}, [ $branch, [] ];

			my @version_infos = split(/^\b/m, join('',
				map { s/^  // or $error = 1; "$_\n" }
					split("\n", $version_infos)
			));
			$error = 1 unless @version_infos;
			foreach (@version_infos) {
				my ($version, $revision0, $revisionl) = /^\Q$category\E(?:--)?\Q$branch\E--([^\s]+)(?:\n  ([^\s]+)(?: \.\. ([^\s]+))?\n)?$/s;
				unless (defined $version) {
					$error = 1; next CATEGORY_ITEM;
				}
				# TODO: consider $full_listing here
				$revision0 = '' unless defined $revision0;
				$revisionl = '' unless defined $revisionl;
				push @{$infos->[-1]->[1]->[-1]->[1]}, [ $version, $revision0, $revisionl ];
			}
		}
	} continue {
		if ($error) {
			warn "Unexpected abrowse output, skipping:\n$_\n";
			pop @$infos;
			$error = 0;
		}
	}
	return $infos;
}

sub get_revision_changeset ($$;$) {
	my $self = shift;
	my $revision = shift;
	my $dir = defined $_[0]? shift: temp_dir_name("arch-changeset");
	die "get_changeset: incorrect dir ($dir)\n" unless $dir && !-d $dir;

	run_tla("get-changeset", $revision, $dir);
	return Arch::Changeset->new($revision, $dir);
}

sub get_changeset ($;$) {
	my $self = shift;
	my $dir = shift;
	my $full_revision = $self->working_name;
	return $self->get_revision_changeset($full_revision, $dir);
}

sub get_revision_log ($) {
	my $self = shift;
	my $revision = shift || die "get_revision_log: No revision given\n";
	my $message = run_tla("cat-archive-log", $revision);
	die "Can't get log of $revision from archive.\n"
		. "Unexisting revision or system problems.\n"
		unless $message;
	return Arch::Log->new($message);
}

sub get_log ($) {
	my $self = shift;
	die "get_log: no working revision\n" unless defined $self->{revision};
	return $self->get_revision_log($self->working_name);
}

sub get_tree ($;$) {
	my $self = shift;
	my $dir = shift || temp_dir_name("arch-tree");
	die "get_tree: incorrect dir ($dir)\n" unless $dir && !-d $dir;
	my $revision = $self->working_name;

	run_tla("get --silent --no-pristine", $revision, $dir);
	die "Can't get revision $revision from archive.\n"
		. "Unexisting revision or system problems.\n"
		unless -d $dir;
	return $dir;
}

1;
